#include <map>
#include <vector>
#include <string>

#include "Asm6502.h"


// all valid ops with all valid addressing modes, and their opcodes
struct op6502_spec {
    std::string op;
    Asm6502::addr_mode mode;
    int code;
};
static struct op6502_spec op_modes[] = {
    {"adc", Asm6502::IMM, 0x69},
    {"adc", Asm6502::ZPG, 0x65},
    {"adc", Asm6502::ZPX, 0x75},
    {"adc", Asm6502::ABS, 0x6D},
    {"adc", Asm6502::ABX, 0x7D},
    {"adc", Asm6502::ABY, 0x79},
    {"adc", Asm6502::INX, 0x61},
    {"adc", Asm6502::INY, 0x71},
    
    {"and", Asm6502::IMM, 0x29},
    {"and", Asm6502::ZPG, 0x25},
    {"and", Asm6502::ZPX, 0x35},
    {"and", Asm6502::ABS, 0x2D},
    {"and", Asm6502::ABX, 0x3D},
    {"and", Asm6502::ABY, 0x39},
    {"and", Asm6502::INX, 0x21},
    {"and", Asm6502::INY, 0x31},
    
    {"asl", Asm6502::IMP, 0x0A},
    {"asl", Asm6502::ZPG, 0x06},
    {"asl", Asm6502::ZPX, 0x16},
    {"asl", Asm6502::ABS, 0x0E},
    {"asl", Asm6502::ABX, 0x1E},
    
    {"bit", Asm6502::ZPG, 0x24},
    {"bit", Asm6502::ABS, 0x2C},
    
    {"bpl", Asm6502::REL, 0x10},
    {"bmi", Asm6502::REL, 0x30},
    {"bvc", Asm6502::REL, 0x50},
    {"bvs", Asm6502::REL, 0x70},
    {"bcc", Asm6502::REL, 0x90},
    {"bcs", Asm6502::REL, 0xB0},
    {"bne", Asm6502::REL, 0xD0},
    {"beq", Asm6502::REL, 0xF0},
    
    {"brk", Asm6502::IMP, 0x00},
    
    {"clc", Asm6502::IMP, 0x18},
    {"cld", Asm6502::IMP, 0xD8},
    {"cli", Asm6502::IMP, 0x58},
    {"clv", Asm6502::IMP, 0xB8},
    
    {"cmp", Asm6502::IMM, 0xC9},
    {"cmp", Asm6502::ZPG, 0xC5},
    {"cmp", Asm6502::ZPX, 0xD5},
    {"cmp", Asm6502::ABS, 0xCD},
    {"cmp", Asm6502::ABX, 0xDD},
    {"cmp", Asm6502::ABY, 0xD9},
    {"cmp", Asm6502::INX, 0xC1},
    {"cmp", Asm6502::INY, 0xD1},
    
    {"cpx", Asm6502::IMM, 0xE0},
    {"cpx", Asm6502::ZPG, 0xE4},
    {"cpx", Asm6502::ABS, 0xEC},
    
    {"cpy", Asm6502::IMM, 0xC0},
    {"cpy", Asm6502::ZPG, 0xC4},
    {"cpy", Asm6502::ABS, 0xCC},
    
    {"dec", Asm6502::ZPG, 0xC6},
    {"dec", Asm6502::ZPX, 0xD6},
    {"dec", Asm6502::ABS, 0xCE},
    {"dec", Asm6502::ABX, 0xDE},
    
    {"dex", Asm6502::IMP, 0xCA},
    {"dey", Asm6502::IMP, 0x88},
    
    {"eor", Asm6502::IMM, 0x49},
    {"eor", Asm6502::ZPG, 0x45},
    {"eor", Asm6502::ZPX, 0x55},
    {"eor", Asm6502::ABS, 0x4D},
    {"eor", Asm6502::ABX, 0x5D},
    {"eor", Asm6502::ABY, 0x59},
    {"eor", Asm6502::INX, 0x41},
    {"eor", Asm6502::INY, 0x51},
    
    {"inc", Asm6502::ZPG, 0xE6},
    {"inc", Asm6502::ZPX, 0xF6},
    {"inc", Asm6502::ABS, 0xEE},
    {"inc", Asm6502::ABX, 0xFE},
    
    {"inx", Asm6502::IMP, 0xE8},
    {"iny", Asm6502::IMP, 0xC8},
    
    {"jmp", Asm6502::ABS, 0x4C},
    {"jmp", Asm6502::IND, 0x6C},
    
    {"jsr", Asm6502::ABS, 0x20},
    
    {"lda", Asm6502::IMM, 0xA9},
    {"lda", Asm6502::ZPG, 0xA5},
    {"lda", Asm6502::ZPX, 0xB5},
    {"lda", Asm6502::ABS, 0xAD},
    {"lda", Asm6502::ABX, 0xBD},
    {"lda", Asm6502::ABY, 0xB9},
    {"lda", Asm6502::INX, 0xA1},
    {"lda", Asm6502::INY, 0xB1},
    
    {"ldx", Asm6502::IMM, 0xA2},
    {"ldx", Asm6502::ZPG, 0xA6},
    {"ldx", Asm6502::ZPY, 0xB6},
    {"ldx", Asm6502::ABS, 0xAE},
    {"ldx", Asm6502::ABY, 0xBE},
    
    {"ldy", Asm6502::IMM, 0xA0},
    {"ldy", Asm6502::ZPG, 0xA4},
    {"ldy", Asm6502::ZPX, 0xB4},
    {"ldy", Asm6502::ABS, 0xAC},
    {"ldy", Asm6502::ABX, 0xBC},
    
    {"lsr", Asm6502::IMP, 0x4A},
    {"lsr", Asm6502::ZPG, 0x46},
    {"lsr", Asm6502::ZPX, 0x56},
    {"lsr", Asm6502::ABS, 0x4E},
    {"lsr", Asm6502::ABX, 0x5E},
    
    {"nop", Asm6502::IMP, 0xEA},
    
    {"ora", Asm6502::IMM, 0x09},
    {"ora", Asm6502::ZPG, 0x05},
    {"ora", Asm6502::ZPX, 0x15},
    {"ora", Asm6502::ABS, 0x0D},
    {"ora", Asm6502::ABX, 0x1D},
    {"ora", Asm6502::ABY, 0x19},
    {"ora", Asm6502::INX, 0x01},
    {"ora", Asm6502::INY, 0x11},
    
    {"pha", Asm6502::IMP, 0x48},
    {"php", Asm6502::IMP, 0x08},
    {"pla", Asm6502::IMP, 0x68},
    {"plp", Asm6502::IMP, 0x28},
    
    {"rol", Asm6502::IMP, 0x2A},
    {"rol", Asm6502::ZPG, 0x26},
    {"rol", Asm6502::ZPX, 0x36},
    {"rol", Asm6502::ABS, 0x2E},
    {"rol", Asm6502::ABX, 0x3E},
    
    {"ror", Asm6502::IMP, 0x6A},
    {"ror", Asm6502::ZPG, 0x66},
    {"ror", Asm6502::ZPX, 0x76},
    {"ror", Asm6502::ABS, 0x6E},
    {"ror", Asm6502::ABX, 0x7E},
    
    {"rti", Asm6502::IMP, 0x40},
    
    {"rts", Asm6502::IMP, 0x60},
    
    {"sbc", Asm6502::IMM, 0xE9},
    {"sbc", Asm6502::ZPG, 0xE5},
    {"sbc", Asm6502::ZPX, 0xF5},
    {"sbc", Asm6502::ABS, 0xED},
    {"sbc", Asm6502::ABX, 0xFD},
    {"sbc", Asm6502::ABY, 0xF9},
    {"sbc", Asm6502::INX, 0xE1},
    {"sbc", Asm6502::INY, 0xF1},
    
    {"sec", Asm6502::IMP, 0x38},
    {"sed", Asm6502::IMP, 0xF8},
    {"sei", Asm6502::IMP, 0x78},
    
    {"sta", Asm6502::ZPG, 0x85},
    {"sta", Asm6502::ZPX, 0x95},
    {"sta", Asm6502::ABS, 0x8D},
    {"sta", Asm6502::ABX, 0x9D},
    {"sta", Asm6502::ABY, 0x99},
    {"sta", Asm6502::INX, 0x81},
    {"sta", Asm6502::INY, 0x91},
    
    {"stx", Asm6502::ZPG, 0x86},
    {"stx", Asm6502::ZPY, 0x96},
    {"stx", Asm6502::ABS, 0x8E},
    
    {"sty", Asm6502::ZPG, 0x84},
    {"sty", Asm6502::ZPX, 0x94},
    {"sty", Asm6502::ABS, 0x8C},
    
    {"tax", Asm6502::IMP, 0xAA},
    {"txa", Asm6502::IMP, 0x8A},
    {"tay", Asm6502::IMP, 0xA8},
    {"tya", Asm6502::IMP, 0x98},
    {"txs", Asm6502::IMP, 0x9A},
    {"tsx", Asm6502::IMP, 0xBA}
};


// initialize the addressing mode map
void Asm6502::init_addr_mode_map(void) {
    int op_modes_len = sizeof(op_modes) / sizeof(op_modes[0]);
    
    for (int i = 0; i < op_modes_len; ++i) {
        m_op_map[op_modes[i].op][op_modes[i].mode] = op_modes[i].code;
    }
}


Asm6502::Asm6502(const char *name, const char **lines) : Assembler(name, lines) {
    init_addr_mode_map();
}


Asm6502::Asm6502(const char *name, std::vector<std::string>& lines) : Assembler(name, lines) {
    init_addr_mode_map();
}


/*
// get a reference to the map of addressing modes and opcodes for op
std::map<Asm6502::addr_mode, int>& Asm6502::get_mode_map(std::string& op) {
    if (! valid_op(op)) {
        throw AsmError("invalid op");
    }
    return m_op_map[op];
}


// get a vector of valid addressing modes for op
std::vector<Asm6502::addr_mode> Asm6502::get_op_modes(std::string& op) {
    std::vector<addr_mode> amlist;
    std::map<addr_mode, int>::iterator it;
    std::map<addr_mode, int> modes = get_mode_map(op);
    
    for (it = modes.begin(); it != modes.end(); ++it) {
        amlist.push_back(it->first);
    }
    return amlist;
}
*/

// get the opcode given an op and its addressing mode
int Asm6502::get_op_code(std::string& op, addr_mode mode) {
    if (! valid_mode(op, mode)) {
        throw AsmError("invalid op mode");
    }
    return m_op_map[op][mode];
}


// check if op is a valid mnemonic
bool Asm6502::valid_op(std::string& op) {
    return (m_op_map.count(op) > 0);
}


// check if mode is a valid addressing mode for op
bool Asm6502::valid_mode(std::string& op, addr_mode mode) {
    if (! valid_op(op)) {
        throw AsmError("invalid op");
    }
    return (m_op_map[op].count(mode) > 0);
}

static int mode_size_tab[] = {
    3, // ABS  Absolute    $1234
    3, // ABX  Absolute,X  $1234,x
    3, // ABY  Absolute,y  $1234,y
    2, // IMM  Immediate   #$12
    1, // IMP  Implied
    3, // IND  Indirect    ($1234)
    2, // INX  Indirect,X  ($12,x)
    2, // INY  Indirect,Y  ($12),y
    2, // REL  Relative    $1234
    2, // ZPG  Zeropage    $12
    2, // ZPX  Zeropage,X  $12,x
    2, // ZPY  Zeropage,y  $12,y
};

int Asm6502::get_mode_size(addr_mode mode) {
    return mode_size_tab[mode];
}
