#include "Asm6502.h"


static bool match_SPACE(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    return token.get_type() == Assembler::Token::SPACE;
}

static bool match_HASH(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    return token.get_type() == Assembler::Token::HASH;
}

static bool match_COMMA(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    return token.get_type() == Assembler::Token::COMMA;
}

static bool match_LPAREN(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    return token.get_type() == Assembler::Token::LPAREN;
}

static bool match_RPAREN(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    return token.get_type() == Assembler::Token::RPAREN;
}

static bool match_EOL(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    return token.get_type() == Assembler::Token::EOL;
}

static bool match_addr(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    if (token.get_type() == Assembler::Token::UINT16) {
        return true;
    } else if (token.get_type() == Assembler::Token::UINT8) {
        return true;
    } else if (token.get_type() == Assembler::Token::STRING) {
        return true;
    } else {
        return false;
    }
}

static bool match_byte(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    if (token.get_type() == Assembler::Token::UINT8) {
        return true;
    } else if (token.get_type() == Assembler::Token::STRING &&
               labeldict.is_defined(token.get_str()) &&
               labeldict.get_size(token.get_str()) == 8) {
        return true;
    } else if (token.get_type() == Assembler::Token::LTSTRING) {
        return true;
    } else if (token.get_type() == Assembler::Token::GTSTRING) {
        return true;
    } else {
        return false;
    }
}

static bool match_x(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    if (token.get_type() != Assembler::Token::STRING) {
        return false;
    }
    if (token.get_str() != "x") {
        return false;
    }
    return true;
}

static bool match_y(Assembler::Token& token, Assembler::LabelDict& labeldict) {
    if (token.get_type() != Assembler::Token::STRING) {
        return false;
    }
    if (token.get_str() != "y") {
        return false;
    }
    return true;
}


static const int maxmatches = 7;

struct mode_6502_match {
    Asm6502::addr_mode mode;
    int nmatches;
    int argpos;
    bool (*match[maxmatches])(Assembler::Token&, Assembler::LabelDict&);
};

// should allow expressions instead of just a single token
static struct mode_6502_match mode_6502_match_tab[] = {
    // try zeropage addressing before absolute, as labels can be cast
    {Asm6502::ZPG, 3, 1, &match_SPACE, &match_byte,   &match_EOL,   NULL,          NULL,         NULL,           NULL      },
    {Asm6502::ZPX, 5, 1, &match_SPACE, &match_byte,   &match_COMMA, &match_x,      &match_EOL,   NULL,           NULL      },
    {Asm6502::ZPY, 5, 1, &match_SPACE, &match_byte,   &match_COMMA, &match_y,      &match_EOL,   NULL,           NULL      },
    {Asm6502::ABS, 3, 1, &match_SPACE, &match_addr,   &match_EOL,   NULL,          NULL,         NULL,           NULL      },
    {Asm6502::ABX, 5, 1, &match_SPACE, &match_addr,   &match_COMMA, &match_x,      &match_EOL,   NULL,           NULL      },
    {Asm6502::ABY, 5, 1, &match_SPACE, &match_addr,   &match_COMMA, &match_y,      &match_EOL,   NULL,           NULL      },
    {Asm6502::IMM, 4, 2, &match_SPACE, &match_HASH,   &match_byte,  &match_EOL,    NULL,         NULL,           NULL      },
    {Asm6502::IMP, 1, 0, &match_EOL,   NULL,          NULL,         NULL,          NULL,         NULL,           NULL      },
    {Asm6502::IND, 5, 2, &match_SPACE, &match_LPAREN, &match_addr,  &match_RPAREN, &match_EOL,   NULL,           NULL      },
    {Asm6502::INX, 7, 2, &match_SPACE, &match_LPAREN, &match_byte,  &match_COMMA,  &match_x,     &match_RPAREN,  &match_EOL},
    {Asm6502::INY, 7, 2, &match_SPACE, &match_LPAREN, &match_byte,  &match_RPAREN, &match_COMMA, &match_y,       &match_EOL},
    {Asm6502::REL, 3, 1, &match_SPACE, &match_addr,   &match_EOL,   NULL,          NULL,         NULL,           NULL      }
};


// given the current opcode string, which is verified to be a legal op,
// determine which addressing mode is being used given a token list of
// arguments
Asm6502::addr_mode Asm6502::determine_addrmode(std::string& op, Assembler::TokenList& tlist) {
    Assembler::TokenList tokens;
    struct mode_6502_match m;
    int nmodes = sizeof(mode_6502_match_tab) / sizeof(mode_6502_match_tab[0]);
    bool matching;
    int i;
    
    // try to parse each valid addressing mode until one matches
    for (int mi = 0; mi < nmodes; ++mi) {
        m = mode_6502_match_tab[mi];
        if (valid_mode(op, m.mode)) {
            i = 0;
            do {
                matching = m.match[i](tlist[i], labeldict);
                ++i;
            } while (matching && i < m.nmatches);
            if (matching) {
                return m.mode;
            }
        }
    }
    throw "invalid addressing mode";
}


// get the token that contains the opcode argument
// FIXME: should return a token list of the whole expression
Assembler::Token Asm6502::get_argument_token(Asm6502::addr_mode mode, Assembler::TokenList& tlist) {
    int nmodes = sizeof(mode_6502_match_tab) / sizeof(mode_6502_match_tab[0]);
    struct mode_6502_match m;

    if (mode == Asm6502::IMP) {
        throw "implied mode has no argument token";
    }
    // FIXME: this should be a quick map lookup instead
    for (int mi = 0; mi < nmodes; ++mi) {
        m = mode_6502_match_tab[mi];
        if (m.mode == mode) {
            if (m.argpos) {
                return tlist[m.argpos];
            } else {
                throw "implied mode has no argument token";
            }
        }
    }
    throw "bad address mode";
}
