#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <cassert>
#include <cctype>

#include "Assembler.h"


// Tokenizer help function
//
// tokenize_X calls count_match_length with match_X as a callback

static size_t count_match_length(std::string& str, bool (*match_callback)(size_t at, char c)) {
    size_t i = 0;
    
    while (i < str.length() && match_callback(i, str.at(i))) {
        ++i;
    }
    
    assert(i > 0);
    return i;
}


static bool match_whitespace(size_t at, char c) {
    return isspace(c) != 0;
}

static Assembler::Token tokenize_whitespace(std::string& line) {
    size_t len;
    
    len = count_match_length(line, &match_whitespace);
    line.erase(0, len);
    
    return Assembler::Token(Assembler::Token::SPACE);
}


static std::string cut_matching_string(size_t saveoffset, std::string& line, bool (*match_callback)(size_t at, char c)) {
    size_t len;
    std::string str;
    
    len = count_match_length(line, match_callback);
    str = line.substr(saveoffset, len);
    line.erase(0, len);
    
    return str;
}


static bool match_string(size_t at, char c) {
    if (at == 0 && (isalpha(c) || c == '_')) {
        return true;
    } else if (at > 0 && (isalnum(c) || c == '_')) {
        return true;
    } else {
        return false;
    }
}


static bool match_prefix_string(char prefix, size_t at, char c) {
    if (at == 0 && c == prefix) {
        return true;
    } else {
        return match_string(at - 1, c);
    }
}


static Assembler::Token tokenize_string(std::string& line) {
    return Assembler::Token(Assembler::Token::STRING, cut_matching_string(0, line, &match_string));
}


static bool match_dotstring(size_t at, char c) {
    return match_prefix_string('.', at, c);
}

static Assembler::Token tokenize_dotstring(std::string& line) {
    return Assembler::Token(Assembler::Token::DOTSTRING, cut_matching_string(0, line, &match_dotstring));
}


static bool match_ltstring(size_t at, char c) {
    return match_prefix_string('<', at, c);
}

static Assembler::Token tokenize_ltstring(std::string& line) {
    return Assembler::Token(Assembler::Token::LTSTRING, cut_matching_string(1, line, &match_ltstring));
}


static bool match_gtstring(size_t at, char c) {
    return match_prefix_string('>', at, c);
}

static Assembler::Token tokenize_gtstring(std::string& line) {
    return Assembler::Token(Assembler::Token::GTSTRING, cut_matching_string(1, line, &match_gtstring));
}


// quoted strings
static Assembler::Token tokenize_quotstring(std::string& line) {
    std::string str;
    size_t i = 1;
    char c;
    
    if (line.length() < 3) {
        throw "malformed quoted string";
    }
    
    while (i < line.length()) {
        c = line.at(i);
        if (c == '"') {
            break;
        }
        if (c == '\\') {
            if (i + 1 < line.length()) {
                ++i;
                c = line.at(i);
                str.push_back(c);
            } else {
                throw "malformed quoted string";
            }
        } else {
            str.push_back(c);
        }
        ++i;
    }
    /*
    // this is why I should comment code better. wtf does this check do?
    if (i == line.length()) {
        throw "malformed quoted string";
    }
    */
    line.erase(0, i);
    
    if (str.length() < 1) {
        throw "malformed quoted string";
    }
    
    return Assembler::Token(Assembler::Token::QUOTSTRING, str);
}


static bool match_decnum(size_t at, char c) {
    return isdigit(c) != 0;
}

static Assembler::Token tokenize_decnum(std::string& line) {
    size_t len;
    int decval = 0;
    std::stringstream ss;
    
    len = count_match_length(line, &match_decnum);
    if (len >= 1 && len <= 5) {
        ss << line.substr(0, len);
        ss >> decval;
    } else {
        throw "malformed decimal number";
    }
    line.erase(0, len);
    
    if (decval <= 255) {
        return Assembler::Token(Assembler::Token::UINT8, decval);
    } else {
        return Assembler::Token(Assembler::Token::UINT16, decval);
    }
}


static bool match_hexnum(size_t at, char c) {
    if (at == 0 && c == '$') {
        return true;
    } else if (at > 0 && isxdigit(c)) {
        return true;
    } else {
        return false;
    }
}

static Assembler::Token tokenize_hexnum(std::string& line) {
    size_t len;
    int hexval = 0;
    std::stringstream ss;
    
    len = count_match_length(line, &match_hexnum);
    if (len < 2) {
        throw "malformed hexadecimal number";
    }
    ss << std::hex << line.substr(1, len - 1);
    ss >> hexval;
    line.erase(0, len);
    
    if (len <= 3) {
        return Assembler::Token(Assembler::Token::UINT8, hexval);
    } else {
        return Assembler::Token(Assembler::Token::UINT16, hexval);
    }
}


static bool match_binnum(size_t at, char c) {
    if (at == 0 && c == '%') {
        return true;
    } else if (at > 0 && (c == '0' || c == '1')) {
        return true;
    } else {
        return false;
    }
}

static int bintodec(const std::string& str) {
    int bin = 0;
    std::string::const_iterator si;
    
    for (si = str.begin(); si != str.end(); ++si) {
        bin <<= 1;
        bin |= *si & 1;
    }
    return bin;
}

static Assembler::Token tokenize_binnum(std::string& line) {
    size_t len;
    int binval = 0;
    
    len = count_match_length(line, &match_binnum);
    if (len == 9) {
        binval = bintodec(line.substr(1, len - 1));
    } else {
        throw "malformed binary number";
    }
    line.erase(0, len);
    
    return Assembler::Token(Assembler::Token::UINT8, binval);
}


static Assembler::Token tokenize_hash(std::string& line) {
    line.erase(0, 1);
    return Assembler::Token(Assembler::Token::HASH);
}


static Assembler::Token tokenize_colon(std::string& line) {
    line.erase(0, 1);
    return Assembler::Token(Assembler::Token::COLON);
}


static Assembler::Token tokenize_equal(std::string& line) {
    line.erase(0, 1);
    return Assembler::Token(Assembler::Token::EQUAL);
}


static Assembler::Token tokenize_comma(std::string& line) {
    line.erase(0, 1);
    return Assembler::Token(Assembler::Token::COMMA);
}


static Assembler::Token tokenize_lparen(std::string& line) {
    line.erase(0, 1);
    return Assembler::Token(Assembler::Token::LPAREN);
}


static Assembler::Token tokenize_rparen(std::string& line) {
    line.erase(0, 1);
    return Assembler::Token(Assembler::Token::RPAREN);
}


Assembler::TokenList Assembler::tokenize_line(std::string& line) {
    char c;
    Assembler::TokenList token_list;
    
    while (line.length()) {
#ifndef NDEBUG
        size_t length_before = line.length(); // debug: make sure we don't get stuck in a loop
#endif
        c = line.at(0);

        if (isspace(c)) {               // whitespace
            token_list.push_back(tokenize_whitespace(line));
        } else if (isalpha(c)) {        // label or opcode
            token_list.push_back(tokenize_string(line));
        } else if (c == '"') {          // quoted string
            token_list.push_back(tokenize_quotstring(line));
        } else if (c == '.') {          // pseudo-op
            token_list.push_back(tokenize_dotstring(line));
        } else if (c == '<') {          // lo-byte of label
            token_list.push_back(tokenize_ltstring(line));
        } else if (c == '>') {          // hi-byte of label
            token_list.push_back(tokenize_gtstring(line));
        } else if (isdigit(c)) {        // decimal number
            token_list.push_back(tokenize_decnum(line));
        } else if (c == '$') {          // hex number
            token_list.push_back(tokenize_hexnum(line));
        } else if (c == '%') {          // binary number
            token_list.push_back(tokenize_binnum(line));
        } else if (c == '#') {          // immediate arg
            token_list.push_back(tokenize_hash(line));
        } else if (c == ':') {          // label def
            token_list.push_back(tokenize_colon(line));
        } else if (c == '=') {          // label assignment
            token_list.push_back(tokenize_equal(line));
        } else if (c == ',') {          // comma
            token_list.push_back(tokenize_comma(line));
        } else if (c == '(') {          // indirect addressing
            token_list.push_back(tokenize_lparen(line));
        } else if (c == ')') {          // indirect addressing
            token_list.push_back(tokenize_rparen(line));
        } else if (c == ';') {          // comment
            line.clear();
        } else {                        // unrecognized character
            throw "unrecognized character";
        }

#ifndef NDEBUG
        if (length_before == line.length()) {
            throw "infinite tokenizer loop";
        }
#endif
    }
    
    // trim whitespace at the end
    while (token_list.size() && token_list.back().get_type() == Assembler::Token::SPACE) {
        token_list.pop_back();
    }
    
    token_list.push_back(Assembler::Token(Assembler::Token::EOL));
    
    return token_list;
}
