#include "output.h"
#include "Assembler.h"


//#include "debug.h"


/*

The assembler takes the source as a NULL terminated array of lines, which are
scanned in two passes:

In the first pass:
* Each line is passed to the tokenizer, which returns a list of tokens.
* The tokens are passed to the parser, which returns an assembly chunk, that
  is saved for pass two.
* The parser also defines all labels it comes across.

In the second pass:
* Each assembly chunk is examined, and any label references are evaluated.
* The chunk is serialized to bytes, and saved to a ram image.

*/


Assembler::Assembler(const char *name, const char **lines) {
    verbose_out() << "Assembling " << name << std::endl;
    m_name = name;
    m_source.clear();
    for (const char **sl = lines; *sl; ++sl) {
        m_source.push_back(*sl);
    }
}


Assembler::Assembler(const char *name, std::vector<std::string>& lines) {
    verbose_out() << "Assembling " << name << std::endl;
    m_name = name;
    m_source = lines;
}


void Assembler::append_chunk(Chunk chunk) {
    m_chunk_list.push_back(chunk);
    m_curr_addr += (addr_t) chunk.get_size();
}


void Assembler::assemble(addr_t start_addr, Memory& memory) {
    Chunk chunk;
    TokenList tlist;
    std::vector<std::string>::iterator si;
    std::vector<Chunk>::iterator ci;
    std::vector<int> data;
    
    m_curr_line = 1;
    m_chunk_list.clear();
    
    try {
        
        // first pass: tokenize, parse, and save chunks
        m_curr_addr = start_addr;
        for (si = m_source.begin(); si != m_source.end(); ++si) {
            debug_out() << "tokenizing line " << m_curr_line << ": " << *si << std::endl;
            tlist = tokenize_line(*si);
#ifdef DEBUG_H
            debug_out() << format_token_list(tlist) << std::endl;
#endif
            append_chunk(parse_line(tlist));
            debug_out() << "added chunk with " << m_chunk_list.back().get_size() << " bytes" << std::endl;
            ++m_curr_line;
        }
        
        // second pass: serialize each chunk to bytes, expanding labels
        for (ci = m_chunk_list.begin(); ci != m_chunk_list.end(); ++ci) {
            debug_out() << "getting chunk" << std::endl;
            chunk = *ci;
            m_curr_addr = chunk.get_addr();
            m_curr_line = chunk.get_line();
            debug_out() << "serializing chunk on line " << m_curr_line << " addr " << format_hex(m_curr_addr, 4) << std::endl;
            data = chunk.serialize_to_bytes(labeldict);
            if (data.size()) {
            debug_out() << "emitting " << data.size() << " bytes of data" << std::endl;
                for (int i = 0; i < (int) data.size(); ++i) {
                    memory.set(m_curr_addr + i, (byte_t) data[i]);
                }
            }
#ifdef DEBUG_H
            debug_print_data(m_curr_line, m_curr_addr, data);
#endif
        }
    }
    catch (const char *msg) {
        throw AsmError(m_name, m_curr_line, msg);
    }
}
