#ifndef ASSEMBLER_H
#define ASSEMBLER_H


#include <map>
#include <vector>
#include <string>
#include <deque>

#include "types.h"
#include "Memory.h"
#include "AsmError.h"


class Assembler {
    
public:
    
    Assembler(const char *name, const char **lines);
    Assembler(const char *name, std::vector<std::string>& lines);
    //Assembler(char *name, ifstream& file);
    virtual ~Assembler() {};
    
    void assemble(addr_t start_addr, Memory& memory);
    
    // Label definitions go in a label dictionary
    class LabelDict {
    public:
        void define(std::string& name, int size, addr_t addr);
        void define(const char *name, int size, addr_t addr);
        bool is_defined(std::string& name);
        int get_size(std::string& name);
        addr_t get_addr(std::string& name);
        addr_t get_addr(const char *name);
        void clear(void);
        std::vector<std::string> get_label_list(void);
    private:
        std::map<std::string,addr_t> m_lab_addr;
        std::map<std::string,addr_t> m_lab_size;
    };
    
    // The source is tokenized to Tokens, stored in TokenLists
    class Token {
    public:
        enum token_type {
            NONE,
            
            SPACE,
            STRING,
            QUOTSTRING,
            DOTSTRING,
            LTSTRING,
            GTSTRING,
            UINT8,
            UINT16,
            HASH,
            COLON,
            EQUAL,
            COMMA,
            LPAREN,
            RPAREN,
            
            EOL
        };
        Token()                                  : m_type(NONE),  m_number(0),   m_string("")  {};
        Token(token_type ttype)                  : m_type(ttype), m_number(0),   m_string("")  {};
        Token(token_type ttype, int num)         : m_type(ttype), m_number(num), m_string("")  {};
        Token(token_type ttype, std::string str) : m_type(ttype), m_number(0),   m_string(str) {};
        token_type get_type(void) { return m_type; };
        int get_num(void) { return m_number; };
        std::string& get_str(void) { return m_string; };
        bool is_numeric(void);
    private:
        token_type m_type;
        int m_number;
        std::string m_string;
    };

    class TokenList : public std::deque<Token> {
    public:
        Token shift(void);
    };

    // The parser takes tokens and emits assembly chunks in the first pass
    class Chunk {
    public:
        enum ref_type {
            NONE,
            LABEL_REF,      // 16-bit label reference
            REL2_LABEL_REF, // relative branch reference from m_addr + 2
            LSB_LABEL_REF,  // lsb of label reference
            MSB_LABEL_REF   // msb of label reference
        };
        Chunk();
        Chunk(int line, addr_t addr);
        Chunk(int line, addr_t addr, int data);
        Chunk(int line, addr_t addr, std::vector<int>& vector_data);
        Chunk(int line, addr_t addr, int opcode, ref_type rt, std::string& label);
        int get_line(void) { return m_line; };
        int get_addr(void) { return m_addr; };
        size_t get_size(void) { return m_size; };
        std::vector<int> serialize_to_bytes(LabelDict& labeldict);
    private:
        int m_line;
        int m_addr;
        size_t m_size;
        std::vector<int> m_data;
        ref_type m_ref_type;
        std::string m_label;
    };

    // parser
    void parse_labeldef(Assembler::Token label, Assembler::TokenList& tlist);
    virtual Chunk parse_opcode(std::string& op, Assembler::TokenList& tlist) = 0;
    Chunk parse_command(std::string& cmd, Assembler::TokenList& tlist);
    Chunk parse_instruction(Assembler::TokenList& tlist);
    Chunk parse_line(Assembler::TokenList& tlist);
    Assembler::Chunk parse_command_org(Assembler::TokenList& tlist);
    Assembler::Chunk parse_command_align(Assembler::TokenList& tlist);
    Assembler::Chunk parse_command_byte(Assembler::TokenList& tlist);
    Assembler::Chunk parse_command_word(Assembler::TokenList& tlist);
    
    // tokenizer
    TokenList tokenize_line(std::string& line);
    
    void append_chunk(Chunk chunk);
    
    const char *m_name;
    int m_curr_line;
    addr_t m_curr_addr;
    std::vector<std::string> m_source;
    
    std::vector<Chunk> m_chunk_list;
    
    LabelDict labeldict;
};


#endif
