#ifndef COMPRESSOR_H
#define COMPRESSOR_H


#include <memory>
#include <stdexcept>

#include "Memory.h"


class CompressorException : public std::runtime_error {
public:
    CompressorException(const std::string& what_arg) : std::runtime_error(what_arg) {};
};


class Compressor {
public:
    typedef enum compression_type {
        NONE,
        LZSS,
        PUCRUNCH,
        EXOMIZER
    };
    virtual ~Compressor() {};
    // compress and add decompressor code, return compressed length
    virtual Memory compress(Memory& mem, addr_t start, addr_t end, addr_t jump_address) = 0;
    // add basic sys line
    void add_sys(Memory& mem);
    // return name of compressor
    virtual const char *get_name(void) = 0;
};


class CompressorFactory {
public:
    std::auto_ptr<Compressor> get_compressor(Compressor::compression_type ct);
};


// NoneCompressor just adds a basic SYS line
class NoneCompressor : public Compressor {
public:
    Memory compress(Memory& mem, addr_t start, addr_t end, addr_t jump_address);
    const char *get_name(void) { return "None"; };
};


// LZSSCompressor uses the built in LZSS compression algorithm
class LZSSCompressor : public Compressor {
public:
    Memory compress(Memory& mem, addr_t start, addr_t end, addr_t jump_address);
    const char *get_name(void) { return "LZSS"; };
};


// PucrunchCompressor calls the external pucrunch compressor
class PucrunchCompressor : public Compressor {
public:
    Memory compress(Memory& mem, addr_t start, addr_t end, addr_t jump_address);
    const char *get_name(void) { return "Pucrunch"; };
};


// ExomizerCompressor calls the external exomizer compressor
class ExomizerCompressor : public Compressor {
public:
    Memory compress(Memory& mem, addr_t start, addr_t end, addr_t jump_address);
    const char *get_name(void) { return "Exomizer"; };
};


#endif /* end of include guard: COMPRESSOR_H */
