#include <iostream>
#include <algorithm>
#include <utility>

#include "output.h"
#include "Asm6502.h"
#include "Compressor.h"


//#include "debug.h"


static const char *lzssunpack[] = {
    "; assumes the following predefined labels:",
    ";   end_of_data:     points to the last data byte",
    ";   npages:          the number of pages to copy to top of ram",
    ";   relocated_data:  points to the first data byte after moving",
    ";   destination:     unpack destination address, and jump address",
    "",
    "",
    "src	= $fb",
    "srchi	= $fc",
    "dest	= $fd",
    "desthi	= $fe",
    "back	= $57",
    "backhi	= $58",
    "",
    "init:",
    "	sei",
    "	inc $01",
    "",
    "	; copy data to top of ram",
    "	; FIXME: selfmod would be smaller",
    "	; FIXME: but needs label + 1",
    "	lda #<end_of_data",
    "	sta src",
    "	lda #>end_of_data",
    "	sta srchi",
    "",
    "	lda #$ff",
    "	sta dest",
    "	lda #$ff",
    "	sta desthi",
    "",
    "	ldx #<npages",
    "	ldy #$00",
    "copyloop:",
    "	lda (src),y",
    "	sta (dest),y",
    "	dey",
    "	cpy #$ff",
    "	bne copyloop",
    "	dec srchi",
    "	dec desthi",
    "	dex",
    "	bne copyloop",
    "",
    "	; start unpacking",
    "",
    "	lda #<relocated_data",
    "	sta src",
    "	lda #>relocated_data",
    "	sta srchi",
    "",
    "	lda #<destination",
    "	sta dest",
    "	lda #>destination",
    "	sta desthi",
    "",
    "back_unpack:",
    "	ldy #$00",
    "	jsr bp_read",
    "	cmp #$00",
    "	bmi literal",
    "	beq done",
    "",
    "; back reference, length - 2 in A",
    "	clc",
    "	adc #$02",
    "	tax",
    "",
    "	jsr bp_read		; read offset byte",
    "",
    "	; inc dest len",
    "",
    "	clc			; back = dest - 256 + offset",
    "	adc dest",
    "	sta back",
    "	ldy desthi",
    "	bcs skipbackhi",
    "	dey",
    "skipbackhi:",
    "	sty backhi",
    "",
    "	ldy #$00		; copy forwards to handle overlap",
    "copyback:",
    "	lda (back),y",
    "	sta (dest),y",
    "	iny",
    "	dex",
    "	bne copyback",
    "",
    "	dey			; count - 1 in A",
    "	tya",
    "	pha",
    "	bne incdest",
    "",
    "done:",
    "	dec $01",
    "	cli",
    "	jmp destination",
    "",
    "",
    "; literal run, length -1 in A with MSB set",
    "literal:",
    "	and #$7f",
    "	tay			; use as counter",
    "",
    "	pha",
    "",
    "copyliteral:",
    "	lda (src),y",
    "	sta (dest),y",
    "	dey",
    "	bpl copyliteral",
    "",
    "	pla			; add length to src pointer",
    "	pha",
    "	sec",
    "	adc src",
    "	sta src",
    "	bcc incdest",
    "	inc srchi",
    "incdest:",
    "	pla			; add length to dest pointer",
    "	sec",
    "	adc dest",
    "	sta dest",
    "	bcc skipdesthi",
    "	inc desthi",
    "skipdesthi:",
    "	jmp back_unpack",
    "",
    "",
    "bp_read:",
    "	lda (src),y",
    "	inc src",
    "	bne bp_read_done",
    "	inc srchi",
    "bp_read_done:",
    "	rts",
    "",
    "start_of_data:",
    NULL
};


class LZSSBuffer {
public:
    LZSSBuffer(Memory& mem, addr_t dest)
    : m_mem(mem), m_dest_index(dest), m_buffer_index(0), m_bytes_written(0) {};
    void write_byte(byte_t byte);
    void buffer_literal_byte(byte_t byte);
    void write_literal_buffer(void);
    int literal_buffer_size(void) { return m_buffer_index; };
    int get_bytes_written(void) { return m_bytes_written; };
    
private:
    Memory& m_mem;
    addr_t m_dest_index;
    byte_t m_literal_buffer[128];
    unsigned int m_buffer_index;
    int m_bytes_written;
};


void LZSSBuffer::write_byte(byte_t byte) {
    m_mem.set(m_dest_index, byte);
    ++m_dest_index;
    ++m_bytes_written;
}


void LZSSBuffer::buffer_literal_byte(byte_t byte) {
    if (m_buffer_index >= sizeof(m_literal_buffer) / sizeof(m_literal_buffer[0])) {
        throw CompressorException("LZSS literal buffer overrun");
    }
    m_literal_buffer[m_buffer_index] = byte;
    ++m_buffer_index;
}


void LZSSBuffer::write_literal_buffer(void) {
    if (m_buffer_index) {
        write_byte((byte_t) (127 + m_buffer_index));
        for (unsigned int i = 0; i < m_buffer_index; ++i) {
            write_byte(m_literal_buffer[i]);
        }
    }
    m_buffer_index = 0;
}


static const int max_match_len = 129;
static const int max_backbuf_size = 256;


// return length, offset to the longest string match
static std::pair<int, int> find_longest_match(byte_t *in, byte_t *buf, int max_len, int backbuf_size) {
    int match_len = 0;
    int match_offset = 0;
    int len;
    int offset = backbuf_size;
    
    while (backbuf_size) {
        if (buf[0] == in[0]) {
            len = 1;
            while (len < max_len && buf[len] == in[len]) {
                ++len;
            }
            if (len > match_len) {
                match_len = len;
                match_offset = offset;
            }
        }
        ++buf;
        --offset;
        --backbuf_size;
    }
    
    return std::make_pair(match_len, match_offset);
}


static int lzss_compress(Memory& mem_out, addr_t dest, Memory& mem_in, addr_t start, addr_t end) {
    LZSSBuffer buffer(mem_out, dest);
    byte_t *inptr;
    byte_t *backptr;
    int bytes_left, bytes_done;
    int max_len, backbuf_size;
    int match_len, match_offset;
    std::pair<int, int> match;
    
    inptr = mem_in.get_valid_range_ptr(start, end);
    bytes_left = end - start + 1;
    bytes_done = 0;
    
    // while there are still bytes left to compress
    while (bytes_left) {
        
        // find a match if there are written bytes, and some more input data
        if (bytes_done && bytes_left) {
            max_len = std::min(bytes_left, max_match_len);
            backbuf_size = std::min(bytes_done, max_backbuf_size);
            backptr = inptr - backbuf_size;
            match = find_longest_match(inptr, backptr, max_len, backbuf_size);
            match_len = match.first;
            match_offset = match.second;
        } else {
            match_len = 0;
        }
        
        // if the match is at least 3 bytes, write a back reference
        if (match_len >= 3) {
            buffer.write_literal_buffer();
            buffer.write_byte((byte_t)(match_len - 2));
            buffer.write_byte((byte_t)(256 - match_offset));
            bytes_done += match_len;
            bytes_left -= match_len;
            inptr += match_len;
        }
        // otherwise we write literal bytes
        else {
            buffer.buffer_literal_byte(*inptr);
            if (buffer.literal_buffer_size() >= 128) {
                buffer.write_literal_buffer();
            }
            ++bytes_done;
            --bytes_left;
            ++inptr;
        }
        
    }
    
    // flush any leftover bytes
    buffer.write_literal_buffer();
    
    // write EOF marker
    buffer.write_byte(0);
    
    // return the size of the compressed output
    return buffer.get_bytes_written();
}

//static int lzss_compress(Memory& mem_out, addr_t dest, Memory& mem_in, addr_t start, addr_t end) {


Memory LZSSCompressor::compress(Memory& mem, addr_t start, addr_t end, addr_t jump_address) {
    Memory mem_out;
    addr_t start_of_data;
    int compressed_len;
    
    try {
        //  end_of_data:     points to the last data byte
        //  npages:          the number of pages to copy to top of ram
        //  relocated_data:  points to the first data byte after moving
        //  destination:     unpack destination address, and jump address
        
        // start with a dry run to determine the end of code, and start of data
        Asm6502 dry_assembler("lzssunpack dry run", lzssunpack);
        dry_assembler.labeldict.define("end_of_data", 16, 0xffff);
        dry_assembler.labeldict.define("npages", 8, 0xff);
        dry_assembler.labeldict.define("relocated_data", 16, 0xffff);
        dry_assembler.labeldict.define("destination", 16, jump_address);
        dry_assembler.assemble(0x080d, mem_out);
        
        // LZSS compress the data
        start_of_data = dry_assembler.labeldict.get_addr("start_of_data");
        compressed_len = lzss_compress(mem_out, start_of_data, mem, start, end);

        debug_out() << "start_of_data: " << format_hex(start_of_data, 4) << std::endl;
        debug_out() << "end_of_data: " << format_hex(start_of_data + compressed_len - 1, 4) << std::endl;
        debug_out() << "npages: " << format_hex(std::min(((compressed_len + 255) >> 8) + 1, 0xf7), 4) << std::endl;
        debug_out() << "relocated_data: " << format_hex(0xffff - compressed_len + 1, 4) << std::endl;
        debug_out() << "destination: " << format_hex(jump_address, 4) << std::endl;

        // re-assemble with the proper labels
        Asm6502 assembler("lzssunpack", lzssunpack);
        assembler.labeldict.define("end_of_data", 16, start_of_data + compressed_len - 1);
        assembler.labeldict.define("npages", 8, std::min(((compressed_len + 255) >> 8) + 1, 0xf7));
        assembler.labeldict.define("relocated_data", 16, 0xffff - compressed_len + 1);
        assembler.labeldict.define("destination", 16, jump_address);
        assembler.assemble(0x080d, mem_out);
    }
    catch (AsmError& e) {
        std::cerr << e.what() << std::endl;
        throw CompressorException("couldn't assemble lzssunpack code");
    }
    
    // return compressed image
    return mem_out;
}
