#include <cstring>

#include "Memory.h"


Memory::Memory(byte_t clear_value)
: m_size(MAXMEM) {
    clear(clear_value);
}


void Memory::clear(byte_t value) {
    memset(m_mem, value, m_size);
    set_highest(-1);
    set_lowest(MAXMEM);
}


void Memory::clear(addr_t start, addr_t end, byte_t value) {
    memset(&m_mem[start], value, end - start + 1);
}



void Memory::set(addr_t addr, byte_t value) {
    if (addr < 0 || addr > m_size) {
        throw MemoryException("set address out of range");
    }
    set_hi_lo(addr, addr);
    m_mem[addr] = value;
}


byte_t Memory::get(addr_t addr) {
    if (addr < 0 || addr >= m_size) {
        throw MemoryException("get address out of range");
    }
    return m_mem[addr];
}


void Memory::copy(addr_t dest_addr, addr_t src_addr, int len) {
    if (! is_range_valid(dest_addr, dest_addr + len - 1)) {
        throw MemoryException("copy destination out of range");
    }
    if (! is_range_valid(src_addr, src_addr + len - 1)) {
        throw MemoryException("copy source out of range");
    }
    set_hi_lo(dest_addr, dest_addr + len - 1);
    memmove(&m_mem[dest_addr], &m_mem[src_addr], len);
}


void Memory::copy(addr_t dest_addr, Memory& src_mem, addr_t src_addr, int len) {
    if (! is_range_valid(dest_addr, dest_addr + len - 1)) {
        throw MemoryException("copy destination out of range");
    }
    set_hi_lo(dest_addr, dest_addr + len - 1);
    src_mem.read(&m_mem[dest_addr], src_addr, len);
}

    
void Memory::read(byte_t *dest_addr, addr_t src_addr, int len) {
    if (! is_range_valid(src_addr, src_addr + len - 1)) {
        throw MemoryException("read source out of range");
    }
    memmove(dest_addr, &m_mem[src_addr], len);
}


void Memory::write(addr_t dest_addr, byte_t *src_addr, int len) {
    if (! is_range_valid(dest_addr, dest_addr + len - 1)) {
        throw MemoryException("write destination out of range");
    }
    set_hi_lo(dest_addr, dest_addr + len - 1);
    memmove(&m_mem[dest_addr], src_addr, len);
}


bool Memory::is_addr_valid(addr_t addr) {
    if (addr < 0 || addr >= m_size) {
        return false;
    }
    return true;
}

bool Memory::is_range_valid(addr_t start, addr_t end) {
    return (is_addr_valid(start) && is_addr_valid(end));
}


void Memory::set_hi_lo(addr_t start, addr_t end) {
    if (start < m_lowest) {
        set_lowest(start);
    }
    if (end > m_highest) {
        set_highest(end);
    }
}


void Memory::set_highest(addr_t addr) {
    m_highest = addr;
}


void Memory::set_lowest(addr_t addr) {
    m_lowest = addr;
}


byte_t *Memory::get_valid_range_ptr(addr_t start, addr_t end) {
    if (is_range_valid(start, end)) {
        return &m_mem[start];
    } else {
        return NULL;
    }
}
